#!/usr/bin/python

#
# 'Automatic' NTA management;
#
# Use in conjunction with a DNS resolver that returns the server
# address of this server when encountering a DNSSEC failure.
#
# If this server gets an http request for a domain name it doesn't
# know, it assumes that this was because of a DNSSEC failure.
# In that case, it will ask whether to set an NTA for that domain name.
#
# If the request is for an IP address or one of the known names,
# it will redirect to that.

import argparse
import ConfigParser
import os
import socket
import subprocess
import sys
import time
import web
import config

DEFAULT_REDIRECT_HOST = "http://tjeb.nl"
DEFAULT_REDIRECT_SUFFIX = "/"

# queries wihtout host are always redirected
# otherwise this list is checked,
KNOWN_HOSTS = [
  'localhost',
  config.server_host
]

urls = [
 '/autonta/set_nta/([a-zA-Z0-9.-]+)', 'SetNTA',
 '/autonta/remove_nta/([a-zA-Z0-9.-]+)', 'RemoveNTA',
 '/autonta/ask_nta/([a-zA-Z0-9.-]+)', 'AskNTA',
 '/autonta', 'NTA',
 '/', 'NTA'
]

render = web.template.render(config.template_dir, base='base')


def store_pid(filename):
    pid = os.getpid()
    with open(filename, 'w') as output:
        output.write("%d\n" % pid)

def split_host(host):
    parts = host.split(':')
    if len(parts) > 1:
        return parts[0], parts[1]
    else:
        return host, None

def run_cmd(arguments):
    proc = subprocess.Popen(arguments,
                            stdout = subprocess.PIPE,
                            stderr = subprocess.PIPE)
    stdout, _ = proc.communicate()
    return stdout.split("\n")[:-1]

def is_valid_ipv4_address(address):
    try:
        socket.inet_pton(socket.AF_INET, address)
    except AttributeError:  # no inet_pton here, sorry
        try:
            socket.inet_aton(address)
        except socket.error:
            return False
        return address.count('.') == 3
    except socket.error:  # not a valid address
        return False

    return True

def is_valid_ipv6_address(address):
    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:  # not a valid address
        return False
    return True

def is_valid_ip_address(address):
    return is_valid_ipv4_address(address) or is_valid_ipv6_address(address)

def is_known_host(host):
    for known_host in KNOWN_HOSTS:
        if host.endswith(known_host):
            return True
    return False

class UnboundNTA:
    def __init__(self, unbound_control):
        self.ub_control = unbound_control

    def add_nta(self, host):
        run_cmd([self.ub_control, "insecure_add", host])

    def remove_nta(self, host):
        run_cmd([self.ub_control, "insecure_remove", host])

    def get_ntas(self):
        print("[XX] executing: " + self.ub_control)
        return run_cmd([ self.ub_control, "list_insecure"])

class DnsmasqNTA:
    def __init__(self, nta_file):
        self.nta_file = nta_file

    def read_ntas(self):
        if not os.path.exists(self.nta_file):
            return []

        # how to propagate errors?
        # standard exception type?
        with open(self.nta_file, 'r') as infile:
            lines = infile.readlines()

        result = []
        for line in lines:
            nta = line.strip()
            if nta != '' and not nta.startswith('#'):
                result.append(nta)
        return result

    def write_ntas(self, ntas):
        with open(self.nta_file, 'w') as outfile:
            for nta in ntas:
                outfile.write("%s\n" % nta)

    def add_nta(self, host):
        host = host.strip()
        ntas = self.read_ntas()
        if host not in ntas:
            ntas.append(host)
        ntas.sort()
        self.write_ntas(ntas)

    def remove_nta(self, host):
        host = host.strip()
        ntas = self.read_ntas()
        if host in ntas:
            ntas.remove(host)
        print("[XX] writing ntas: " + str(ntas))
        self.write_ntas(ntas)

    def get_ntas(self):
        return self.read_ntas()

class SetNTA:
    def GET(self, host):
        print("[GET]")
        # TODO: full URI.
        config.backend.add_nta(host)
        return render.nta_set(host)

class RemoveNTA:
    def GET(self, host):
        print("[GET]")
        config.backend.remove_nta(host)
        raise web.seeother("/")

class AskNTA:
    def GET(self, host):
        print("[GET]")
        # make a list of domains to possibly set an NTA for
        if host.endswith('.'):
            host = host[:-1]
        domains = []
        labels = host.split('.')
        lc = len(labels)
        for i in range(lc-1):
            domains.append(".".join(labels[i:lc]))
            
        return render.ask_nta(host, domains)

class NTA:
    def GET(self):
        print("[GET]")
        host = web.ctx.env.get('HTTP_HOST')
        (host, port) = split_host(host)
        print("[XX] host: %s" % host)
        if is_valid_ip_address(host) or is_known_host(host):
            # show NTA list?
            ntas = config.backend.get_ntas()
            return render.nta_list(ntas)
        else:
            if host + "." in get_ntas():
                return render.nta_set(host)
            if port is not None:
                redirect = "http://%s:%s/autonta/ask_nta/%s" % (config.server_host, port, host)
            else:
                redirect = "http://%s/autonta/ask_nta/%s" % (config.server_host, host)
            print("[XX] redirect to %s" % redirect)
            raise web.seeother(redirect)
            
        raise web.seeother(DEFAULT_REDIRECT_HOST + DEFAULT_REDIRECT_SUFFIX)

def main():
    arg_parser = argparse.ArgumentParser(prog="autonta.py")
    arg_parser.add_argument('-c', '--config',
                            default=config.config_file,
                            help='Set configuration file (default %s)' %
                            config.config_file)

    args = arg_parser.parse_args()

    if not os.path.exists(args.config):
        print("Config file '%s' does not exist" % args.config)
        sys.exit(1)
    user_config = ConfigParser.SafeConfigParser()
    user_config.read([args.config])
    if user_config.has_section('unbound'):
        config.backend = UnboundNTA(user_config.get('unbound', 'unbound_control'))
    elif user_config.has_section('dnsmasq'):
        config.backend = DnsmasqNTA(user_config.get('dnsmasq', 'nta_file'))
    else:
        print("Error: no backend specified")
        sys.exit(1)
    
    store_pid(config.pid_file)

    # overwrite cli-args in order to run, we parsed them anyway
    server_port = config.server_port
    server_address = config.server_address

    sys.argv[1] = "%s:%d" % (server_address, server_port)
    app = web.application(urls, globals())
    if config.wsgi:
        web.wsgi.runwsgi = lambda func, addr=None: web.wsgi.runfcgi(func,addr)
    app.run()
    

if __name__ == "__main__":
    main()
